% demo for Arrabit to compute principal eigen-decomposition

clear

% add path
addpath('../examples');
addpath('../src');

if exist('rng','builtin'), rng('default'); end


fprintf('\n\nThe performance of Arrabit may be determined by the efficiency\n');
fprintf('of the matrix-matrix multiplications AX. For example, the sparse\n');
fprintf('matrix-dense vector products using intel MKL often speed up\n');
fprintf('the aglorithm signficantly. For examples on configuring MKL,\n');
fprintf('please refer to Readme.txt\n\n');


SpMV_MKL = 0;

% tolerance
tol = 1e-12;
opts.tol = tol;
opts.idisp = 0;

%% problem name
MatName = ...
   ['1 delsq  ';
    '2 wathen ';
    '3 poisson';
    '4 neumann';
    '5 clement';
    '6 toeppen';
    '7 lesp   '];
%disp(MatName)

sigma = []; Pn = []; Ms = [];
%sigma = input('SIGMA = ','s');
%Pn = input('Problem [1:7]: ');
%Ms = input('   Size [3:7]: ');  

if isempty(sigma), sigma = 'LA'; end
if isempty(Pn), Pn = 1:2; end
if isempty(Ms), Ms = (5:10); end

np = numel(Pn);
nm = numel(Ms);
T1 = zeros(np,nm);
T2 = zeros(np,nm);
T3 = zeros(np,nm);
Ns = zeros(np,nm);


for i = 1:np

    p = Pn(i);
    matname = MatName(p,3:end);
    for j = 1:nm
        
        %% construct matrix A and operator Afun
        m = Ms(j);
        n = (8*m)^2;
        switch p
            case 1, A = delsq(numgrid('L',10*m+2));
            case 2, A = gallery('wathen',5*m,5*m);
            case 3, A = gallery('poisson',(9*m));
            case 4, A = gallery('neumann',n);
            case 5, A = gallery('clement',n,1)/n;
            case 6, A = gallery('toeppen',n,1,-10,0,-10,1);
            case 7, A = gallery('lesp',n); %opts.degs=max(3,Degs);
        end
        mA = max(max(abs(A)));
        A = sparse(A) / mA;
        A = (A + A')/2;        % symmetrize
        n = size(A,1);         % dim of A
        rr = 20;
        k = round(n/rr);       % number of evs to compute
        fprintf(['\n' matname ': (n,k,tol) = (%i,%i,%.e)\n'],n,k,tol)
        Ns(i,j) = n;
        
        if ~SpMV_MKL; Aop = @(X) A*X; else Aop = @(X) sfmult(A,X,1); end

        %% estimate extremers by Gershgorin circles 
        d = full(diag(A)); r = sum(abs(A),2)-abs(d);
        l1 = max(d + r); ln = min(d - r); use_Gersh = 0;
        if strcmp(sigma,'LA') && use_Gersh
            opts.l1 =  l1; opts.ln =  ln;
        end
        if strcmp(sigma,'SA') && use_Gersh
            opts.l1 = -ln; opts.ln = -l1;
        end

        %% MPM
        opts.innsolver = @mpmirr;
        tic, [X,D,flag] = arrabit(Aop,n,k,sigma,opts);
        [ev1,is] = sort(diag(D),'descend');
        X = X(:,is); AX = A*X; D = diag(ev1);
        R = sqrt(sum((AX-X*D).^2))./abs(ev1');
        res1 = max(R); T1(i,j) = toc; 
        fprintf('MPM  time: %f\n',T1(i,j))
        % high accuracy percentage
        hp = round(sum(R < tol)/k*100);
  
        %% GN
        opts.innsolver = @GN_slrp;
        tic, [X,D,flag] = arrabit(Aop,n,k,sigma,opts);
        [ev2,is] = sort(diag(D),'descend');
        X = X(:,is); AX = A*X; D = diag(ev2);
        R = sqrt(sum((AX-X*D).^2))./abs(ev2');
        res2 = max(R); T2(i,j) = toc; 
        fprintf('GN   time: %f\n',T2(i,j))
        % high accuracy percentage
        hp = round(sum(R < tol)/k*100);
        
        %% EIGS
        tic; [X,D] = eigs(A,k,sigma,struct('tol',tol));
        [ev3,is] = sort(diag(D),'descend');
        X = X(:,is); AX = A*X; D = diag(ev3);
        R = sqrt(sum((AX-X*D).^2))./abs(ev3');
        res3 = max(R); T3(i,j) = toc; 
        fprintf('EIGS time: %f\n',T3(i,j))
        
        %% printout
        fprintf('MPM  eigvec Res: %.2e\n',res1)
        fprintf('GN   eigvec Res: %.2e\n',res2)
        fprintf('EIGS eigvec Res: %.2e\n',res3)
        fprintf('MPM-EIGS Max eigenval rel-diff: %.2e\n',...
            norm(ev1(:)./ev3(:)-1,inf))
        fprintf('GN-EIGS  Max eigenval rel-diff: %.2e\n',...
            norm(ev2(:)./ev3(:)-1,inf))
    end
    
    disp(' ')
    disp('    Matrix size (row1) and eigenspace dim (row2):');
    disp([Ns(i,:); round(Ns(i,:)/rr)]);
    disp('    Run time by MPM (row1), GN (row2) and EIGS (row3):');
    disp([T1(i,:); T2(i,:); T3(i,:)])
end

%% plotting
fprintf('\n    Summary: tol = %.2e.  Problem: \n',tol); 
disp(Pn)
fprintf('    Average time by MPM (row1), GN (row2) and EIGS (row3):\n');
disp([mean(T1,1); mean(T2,1); mean(T3,1)])

fig = figure(1);
plot(1:nm,mean(T1,1),'bo-.',1:nm,mean(T2,1),'rs:',...
    1:nm,mean(T3,1),'d--','linewidth',2);
title('Average Time'); set(gca,'fontsize',14);
legend('MPM','GN','EIGS','location','best'); grid on
shg

% save(strcat('results',filesep,'res4eig'), 'Ms', 'Ns', 'MatName', ...
%     'opts', 'T1', 'T2', 'T3');

print(fig, '-djpeg',strcat('results',filesep,'res4eig'));


