function [X,ev,Res,Out] = GN_slrp(Ao,Mfun,X,k,kw,ln,opts)
% Gauss Newton: min 0.5||XX' - A||_F^2
%
% Usage: [X,ev,Res,Out] = GN_slrp(Ao,Mfun,X,opts)
% Input:
%        Ao -- original n by n matrix
%            (numeric or function handle)
%        Mfun -- function handle for a matrix that has
%            the same principal eigenvectors as Ao
%        X -- n by kw starting matrix for a basis
%        opts -- options (see initialize for details)
% Output:
%        X -- computed principal eigenvectors
%             (n by k orthonormal matrix)
%        ev -- computed principal eigenvalues (vector)
%        Res -- residual norm(A*X - X*ev)./abs(ev)
%        Out -- secondary output quantities
% Note:
%        Output eigenpairs may or may not be accurate.
%        Computed residuals will show obtained accuracy.
%
% Copyright 2013 - 2015 Yin Zhang
% Version 0.8, August 30, 2014
% Version 1,   July   28, 2015: Zaiwen Wen

if nargin <= 3;  k  = size(X,2);  end
if nargin <= 4;  kw = size(X,2);  end
if nargin <= 5;  ln = 0;  end
if nargin <= 6;  opts = [];  end

if isnumeric(Ao), A = Ao; Ao = @(X)A*X; end
opts = initialize(opts); Out.it0 = 0;
colnrm = @(X)sqrt(sum(X.^2,1));
nrmcol = @(X)bsxfun(@times,X,1./colnrm(X));

kk = size(X,2);
kQt = size(Qt,2);

tAX = 0;
tstart = tic;
if ~isempty(opts.evmfun); 
    X = bsxfun(@times,X,sqrt(opts.evmfun));
end

optchol1.LT = true; optchol1.TRANSA = false;
optchol2.LT = true; optchol2.TRANSA = true;
rc = 10; flag = 1;
for it1 = 1:maxit1  % main loop
    for it0 = 1:maxit0  % level-0 loop

        T = X'*X;
        Y = X/T;
%         %Y1 = X/T;
%         [La, pp] = chol(T,'lower');
%         if pp == 0
%             %Y = (La'\(La\(X')))';
%             Y = linsolve(La,linsolve(La,X',optchol1),optchol2); Y = Y';
%         else
%             [Y, pp] = qr(X,0);
%             printf('\naa\n');
%         end
%         %norm(Y-Y1,'fro')
        
        t1 = tic; AY = Mfun(Y); tAX = tAX + toc(t1);
        YtAY = Y'*AY;
        %S = YtAY/2 - Ih;
        S = (YtAY - speye(kk))/2;
        X = AY - X*S;
        X = proj2Qker(X);  kk = size(X,2);
    end
    Out.it0 = Out.it0 + maxit0;
    if it1 == maxit1, break; end
    %T = X'*X;
    rcp = rc; rc = rcond(T);
    rc2 = rc/rcp;
    if idisp, fprintf('--- rc: %.2e  rc2: %.2e\n',rc,rc2); end
    if rc  < tol,  flag = 0; break; end
    if rc2 > 0.99, flag = 2; break; end    
end
tSU = toc(tstart);

tstart = tic;
%% augmentation
Z = X; 
if doARR
    for j = 1:nblks-1  %#ok<*AGROW>
        t1 = tic; X = Ao(X)-ln*X; tAX = tAX + toc(t1);
        Z = [Z X];
    end
end
Z = nrmcol(Z);

%% (possibly augmented) RR
if isempty(Qc) || defl < 2
    if ~isempty(Qc), Z = [Qc Z]; end
    [Z,~] = qr(Z,0);
    [X,ev,tt] = RR(Ao,Z);
    X = X(:,1:kw);
    ev = ev(1:kw);    
else %~isempty(Qc) && defl = 2
    Z = proj2Qker(Z);
    [Z,~] = qr(Z,0);
    [X,ev,tt] = RR(Ao,Z);
    %keep those X columns more orth. to Qc
    [~,io] = sort(colnrm(Qc'*X),'ascend');
    io = io(1:(kw-numel(evc)));
    X  = [Qc X(:,io)];
    ev = [evc ev(io)];
    [ev,id] = sort(ev,'descend'); 
    X = X(:,id);
end
RRw = size(Z,2);
%X = X(:,1:kw); 
%ev = ev(1:kw);
tAX = tAX + tt;


%% residual
% Z = Ao(X);
% evs = abs(ev);
% Res = colnrm(Z-bsxfun(@times,X,ev))./max(1,evs);
% %evs = max(mean(evs),evs);
% %Res = colnrm(Z-bsxfun(@times,X,ev))./evs;

t1 = tic; Z = Ao(X(:,1:k)); tAX = tAX + toc(t1);
evs = abs(ev(1:k));
Res = colnrm(Z-bsxfun(@times,X(:,1:k),ev(1:k)))./max(1,evs);
tRR = toc(tstart);

%% output quantities
Out.rc = rc;
Out.it1 = it1;
Out.RRw = RRw;
Out.flag = flag;
Out.tSU = tSU;
Out.tRR = tRR;
Out.tAX = tAX;

%% nested function in GN_slrp %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function opts = initialize(opts)
        if ~isfield(opts,'maxit0'); opts.maxit0 = 5;  end
        if ~isfield(opts,'maxit1'); opts.maxit1 = 10; end
        if ~isfield(opts,'nblks');  opts.nblks = 2;   end
        if ~isfield(opts,'idisp');  opts.idisp = 0;   end
        if ~isfield(opts,'defl');   opts.defl = 1;    end
        if ~isfield(opts,'tol');    opts.tol = 5e-11; end
        if ~isfield(opts,'evc');    opts.evc = [];    end
        if ~isfield(opts,'Qc');     opts.Qc  = [];    end
        if ~isfield(opts,'Qt');     opts.Qt  = [];    end
        if ~isfield(opts,'evmfun'); opts.evmfun  = [];    end
        if ~isfield(opts,'doARR'); opts.doARR = 1;   end
        doARR  = opts.doARR;
        maxit0 = opts.maxit0;     % for level 0 iterations
        maxit1 = opts.maxit1;     % for level 1 iterations
        nblks  = opts.nblks;      % no. of augmented blocks
        idisp  = opts.idisp;      % display control
        defl   = opts.defl;       % deflation control
        tol    = opts.tol;        % tolerance for level 0
        evc    = opts.evc;        % converged eigenvalues
        Qc     = opts.Qc;         % converged eigenvectors
        Qt     = opts.Qt;        % trimmed eigenvectors
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function X = proj2Qker(X)
        % projection to trimmed kernel
        if ~isempty(Qt), X = X - Qt*(Qt'*X); end

        % projection to Qc's kernel
        if ~isempty(Qc), X = X - Qc*(Qc'*X); end
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

end %mpmirr

%% external function %%%%%%%%%%%%%%%%%%%%%%%%%%%
function [U,d,tt] = RR(Afun,X,k)
% Rayleigh-Ritz procedure: [U,d] = RR(Afun,X,k)
% where X is orthonormal, Afun represents the
% action of a real symmetric matrix.  It returns
% Ritz pairs in (U,d) where d is a row vector of
% k Ritz values in decending order.

if isnumeric(Afun)
    A = Afun;
    Afun = @(X)A*X;
end

kc = size(X,2);
if nargin < 3, k = kc;
elseif k > kc, k = kc;
    warning('k set to the number of columns');
end

t1 = tic;
XtAX = X'*Afun(X); tt = toc(t1);
XtAX = (XtAX + XtAX')/2;
[V,D] = eig(XtAX);
[d,p] = sort(real(diag(D)),'descend');
U = X*V(:,p(1:k)); d = d(1:k)';

end
