function [U,D,Out] = arrabit(varargin)
%
% Arrabit Method Outer loop driver for computing
%     k extreme eigenpairs of symmetric matrix A
%
% !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
%
% The performance of Arrabit may be determined by the efficiency
% of the matrix-matrix multiplications AX. For example, the sparse
% matrix-dense vector products using intel MKL often speed up
% the aglorithm signficantly. For examples on configuring MKL,
% please refer to Readme.txt
%
%   http://bicmr.pku.edu.cn/~wenzw/arrabit.html#issue
% !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
%
%
%  Usage: [U,D,Out] = arrabit(A,k,sigma,opts)
%     or [U,D,Out] = arrabit(Afun,n,k,sigma,opts)
% 
% Input:
%        A -- n by n real symmetric matrix (numeric)
%     or Afun,n -- function handle and size of A
%        k -- number of extreme eigenpairs to compute
%    sigma -- 'LA' or 'SA', Largest or Smallest Algebraic
%     opts -- optional parameters
% Output:
%        U -- computed k extreme eigenvectors
%               (n by k column-orthonormal matrix)
%        D -- computed k extreme eigenvalues
%               (k by k sparse matrix)
%      Out -- output information
% 
% Copyright 2013 - 2015 Yin Zhang
% Version 0.8, August 30, 2014
% Version 1,   July   28, 2015: Zaiwen Wen



%% initialize
% argOffset indicates where are inputs k, sigma and opts
A0 = varargin{1};
if isnumeric(A0);
    argOffset = 1; n = size(A0,1);
    if n ~= size(A0,2); error('A must be square'); end
    Ao = @(X)A0*X; clear A0;
else
    argOffset = 0; n = varargin{2};
    Ao = A0; clear A0;
end
% k: the number of desired eigenvalues
if nargin < 3-argOffset
    k = min(n,100);
else k = varargin{3-argOffset};
end

% sigma: largest or smallest
if nargin < 4-argOffset
    sigma = 'LA';
else sigma = varargin{4-argOffset};
end
if strcmpi(sigma(1:2),'LA')
    largest = true;
else largest = false;
    Ao = @(X)-Ao(X);
end
% opts: optional values
if nargin < 5-argOffset
    opts = [];
else opts = varargin{5-argOffset};
end

if n == 0; error('matrix A is empty'); end
if k > n/3; error('k>n/3. k is too large. Try eig!'); end
%if k > n/3; warning('k > n/3. k is too large. Try eig!'); end

if ~isfield(opts,'squared'); opts.squared = 0; end
if ~isfield(opts,'cheby'); opts.cheby = 0;     end
if ~isfield(opts,'maxit'); opts.maxit = [5 10 30];end
if ~isfield(opts,'nblks'); opts.nblks = 2;     end
if ~isfield(opts,'maxblks'); opts.maxblks = min(4,2*opts.nblks); end
if ~isfield(opts,'guard'); opts.guard = 0.1;   end
if ~isfield(opts,'doARR'); opts.doARR = 1;     end
if ~isfield(opts,'idisp'); opts.idisp = 0;     end
if ~isfield(opts,'defl');  opts.defl  = 2;     end
if ~isfield(opts,'deg0');  opts.deg0  = 3;     end
if ~isfield(opts,'maxdeg');opts.maxdeg = 15;   end
if ~isfield(opts,'fixdeg');opts.fixdeg = 0;    end
if ~isfield(opts,'evcut');opts.evcut = 0.9;    end
if ~isfield(opts,'trim');  opts.trim  = 1;     end
if ~isfield(opts,'tol');   opts.tol   = 1e-6;  end
if ~isfield(opts,'tol0');  opts.tol0  = max(opts.tol,1e-6); end
if ~isfield(opts,'tolInt');opts.tolInt= 2;     end
if ~isfield(opts,'ln');    opts.ln   = [];     end
if ~isfield(opts,'l1');    opts.l1   = [];     end
if ~isfield(opts,'lkw');   opts.lkw  = [];     end
if ~isfield(opts,'rgap');  opts.rgap = 0.95;   end
if ~isfield(opts,'matname');   opts.matname  = '';     end
if ~isfield(opts,'mpmsolver');   opts.mpmsolver  = @mpmo;  end
if ~isfield(opts,'innsolver');   opts.innsolver  = @mpmirr;  end
if ~largest && ~isempty(opts.l1)
    opts.ln = -opts.l1; 
end

% X: initial values
if nargin < 6-argOffset
    X = [];
else
    X = varargin{6-argOffset};
end

%% call solver
[U,ev,Out] = feval(opts.mpmsolver,Ao,n,k,opts,X);
if not(largest), ev = -ev; end
D = spdiags(ev(:),0,k,k);
Out.flag = ~Out.rule(1);

end
