/*
 *
 * [x, e, res, m, info, epsout, loop] =
 *      dfeast_syev(A, m0, emin, emax, opts, x0)
 *
 * call dfeast_syev(uplo, n, a, lda, fpm, epsout, loop, emin, emax, ...
 *                  m0, e, x, m, res, info)
 *
 *
 * Compilation in
 *
 * unix/linux:
 * mex -O -largeArrayDims -output dfeast dfeast.cpp
 *
 * Author: Zaiwen Wen
 * wenzw@math.pku.edu.cn
 * 2014-8-20
 *
 * reference:
 *   https://software.intel.com/sites/products/documentation/doclib/mkl_sa/11/mklman/index.htm#GUID-A1F983C8-8B4F-450C-9FEC-96A988EC68A6.htm
 */

#include "mex.h"
#include "mkl.h"

#include <math.h>
#include <string.h> /* needed for memcpy() */


#if !defined(MX_API_VER) || ( MX_API_VER < 0x07030000 )
typedef int mwIndex;
typedef int mwSize;
#endif

#if !defined(MAX)
#define	MAX(A, B)	((A) > (B) ? (A) : (B))
#endif

#if !defined(MIN)
#define	MIN(A, B)	((A) < (B) ? (A) : (B))
#endif

void mexFunction(int nlhs, mxArray *plhs[], int nrhs,
        const mxArray *prhs[])
{
    
    size_t n, i, k;
    mwIndex *A_Ir, *A_Jc;
    MKL_INT ii, nnz, nstep, maxStep, quiet, eigType;
    
    MKL_INT fpm[128];
    
    // MKL representation
    char up[] = {'U'}, lo[] = {'L'}, ff[] = {'F'}, srtID[] = {'D'};
    
    MKL_INT A_n[1], m[1], m0[1], m0p[1], k0, info[1], srt_info[1], loop[1];
    MKL_INT *A_indx, *A_pntr;
    double  *A,  *x, *e, *res, epsout[1];
    double *emin, *emax, *tmp, kFactor, etime;
    bool spC;
    
    if(nrhs < 4){
        mexErrMsgTxt("at least four inputs are needed"); return;
    }
    
    
    feastinit(fpm);
    
    /* set the max number of out steps*/
    maxStep = 10;
    
    /* eigType = 1, search eigenvalues in [emin, emax]
     * eigType = 2, find m0 largest eigenvalues*/
    eigType = 2;
    kFactor = 2.0;
    /* opts */
    if(nrhs >= 5){
        if (mxGetField(prhs[4], 0, "quiet") != NULL ){
            tmp = mxGetPr(mxGetField(prhs[4], 0, "quiet"));
            quiet = (MKL_INT)*tmp;
            if(quiet<=0) fpm[0] = 0; else fpm[0] = 1;
        }
        
        if (mxGetField(prhs[4], 0, "nContour") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "nContour"));
            fpm[1] = (MKL_INT)*tmp;}
        
        if (mxGetField(prhs[4], 0, "dErrTrace") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "dErrTrace"));
            fpm[2] = (MKL_INT)*tmp;}
        
        if (mxGetField(prhs[4], 0, "nLoop") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "nLoop"));
            fpm[3] = (MKL_INT)*tmp;}
        
        if (mxGetField(prhs[4], 0, "InitSubspace") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "InitSubspace"));
            fpm[4] = (MKL_INT)*tmp;}
        
        if (mxGetField(prhs[4], 0, "stopRule") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "stopRule"));
            fpm[5] = (MKL_INT)*tmp;}
        
        if (mxGetField(prhs[4], 0, "sErrTrace") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "sErrTrace"));
            fpm[6] = (MKL_INT)*tmp;}
        
        if (mxGetField(prhs[4], 0, "maxStep") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "maxStep"));
            maxStep = (MKL_INT)*tmp;}
        
        if (mxGetField(prhs[4], 0, "eigType") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "eigType"));
            eigType = (MKL_INT)*tmp;}
        
        if (mxGetField(prhs[4], 0, "kFactor") != NULL ) {
            tmp = mxGetPr(mxGetField(prhs[4], 0, "kFactor"));
            kFactor = *tmp;}
    }
    
    n  = mxGetM(prhs[0]);
    if ( n != mxGetN(prhs[0]))
        mexErrMsgTxt("A must be a square matrix");
    
    tmp  = mxGetPr(prhs[1]); k0 = (MKL_INT) *tmp; /*k = (size_t) *tmp;*/
    emin = mxGetPr(prhs[2]);
    emax = mxGetPr(prhs[3]);
    
    /* set the initial k0*/
    if(eigType == 1) m0[0] = k0;
    else {
        m0[0] = (MKL_INT) (kFactor*k0);
    }
    
    k = (size_t) m0[0];
    
    /* Get the dimensions of the matrix input u. */
    A_n[0] = (MKL_INT) n;
    
    spC = mxIsSparse(prhs[0]);
    if(spC) {
        A    = mxGetPr(prhs[0]);
        A_Ir = mxGetIr(prhs[0]);
        A_Jc = mxGetJc(prhs[0]);
        
        A_indx = (MKL_INT*) mxMalloc(A_Jc[n]*sizeof(MKL_INT));
        A_pntr = (MKL_INT*) mxMalloc((n)*sizeof(MKL_INT));
        
        // MKL needs 1-based indexing for CSR
        for (ii=0; ii < (MKL_INT) A_Jc[n]; ii++)
            A_indx[ii] = (MKL_INT) A_Ir[ii] + 1;
        for (ii=0; ii < A_n[0]+1; ii++)
            A_pntr[ii] = (MKL_INT) A_Jc[ii] + 1;
        
        /*
         * nnz = A_Jc[k];
         * mexPrintf("\n");
         * for(ii = 0; ii < nnz; ii++)
         * mexPrintf("row: %d, val: %e, A_indx: %d\n", A_Ir[ii], A[ii], A_indx[ii]);
         * mexPrintf("\n");
         * for(ii = 0; ii <= k; ii++)
         * mexPrintf("col: %d, Jc: %d, A_pntr: %d\n", ii,  A_Jc[ii], A_pntr[ii]);
         */
        
    }
    else    {
        A    = mxGetPr(prhs[0]);
    }
    
    /*
     * mexPrintf("n: %d, m0: %d, k: %d, ee: [%3.2e, %3.2e]\n", n,  m0[0], k, *emin, *emax);
     */
    
    /*
     * for(ii=0;ii<=20;ii++){
     * mexPrintf("fpm[%d]: %d\n", ii, fpm[ii]);
     * } */
    
    plhs[0] = mxCreateDoubleMatrix(n, k, mxREAL);
    x = mxGetPr(plhs[0]);
    /* copy x0 to x */
    if(nrhs >= 6){
        memcpy(x,mxGetPr(prhs[5]),(n*k)*sizeof(double));
    }
    
    /*  0  1   2    3    4    5       6
     * [x, e, res, m, info, epsout, loop] =
     * dfeast_syev(A, m0, emin, emax, opts, x0) */
    
    plhs[1] = mxCreateDoubleMatrix(k, 1, mxREAL); e = mxGetPr(plhs[1]);
    plhs[2] = mxCreateDoubleMatrix(k, 1, mxREAL); res = mxGetPr(plhs[2]);
    
    m0p[0] = m0[0];
    for(nstep = 1; nstep <= maxStep; nstep++){
        /* call FEAST */
        m0[0] = m0p[0];
        if(quiet >= 1) mexPrintf("step: %2d, interval: [%e, %e], m0p: %5d, ",
                nstep, emin[0], emax[0], m0[0]);
        
        etime = dsecnd();
        if(spC) {
            /* mexPrintf("call sparse \n"); */
            dfeast_scsrev(ff, A_n, A, A_pntr, A_indx, fpm, epsout, loop,
                    emin, emax, m0, e, x, m, res, info);
        }
        else {
            /* mexPrintf("call dense \n"); */
            dfeast_syev(ff, A_n, A, A_n, fpm, epsout, loop,
                    emin, emax, m0, e, x, m, res, info);
        }
        
        etime = dsecnd() - etime ;
        
        if(quiet >= 1 )
            mexPrintf(" m0: %5d, m: %5d, info: %2d, cpu: %f\n",
                    m0[0], m[0], info[0], etime);
        
        if(info[0] == 0) {
            break;
            
            if(m[0] >= k0) break;
            else {
                dlasrt( srtID, m0p, e, srt_info);
                if(quiet>=3){
                    mexPrintf("srt_info: %d\n", srt_info[0]);
                    for(ii=0; ii<m0p[0]; ii++)
                        mexPrintf("e[%d]: %e\n", ii, e[ii]);
                }
                
                fpm[4] = 1;
                emax[0] = MAX(emax[0], e[0]);
                ii = MAX(0, MIN(k0, m0[0])-1);
                emin[0] = e[ii];
                /*mexPrintf(" kk: %d, emin: %e\n", ii, emin[0]); */
                

            }

        }
        else if (info[0] == 3){
            /*if(m[0] >= k0) break;*/
            
            dlasrt( srtID, m0p, e, srt_info);
            if(quiet>=3){
                mexPrintf("srt_info: %d\n", srt_info[0]);
                for(ii=0; ii<m0p[0]; ii++)
                    mexPrintf("e[%d]: %e\n", ii, e[ii]);
            }
            
            fpm[4] = 1;
            emax[0] = MAX(emax[0], e[0]);
            
            ii = MAX(0, MIN(k0, m0[0])-1);
            emin[0] = e[ii];
            /*mexPrintf(" kk: %d, emin: %e\n", ii, emin[0]);*/

            /*
             * if(quiet>=2)
             * mexPrintf("emax: %e, emin: %e\n\n", emax[0], emin[0]);*/
        }
        else break;
    }
    
    if(nlhs >= 4){
        plhs[3] = mxCreateDoubleMatrix(1, 1, mxREAL);
        tmp = mxGetPr(plhs[3]); *tmp = (double) m[0];
    }
    
    if(nlhs >= 5){
        plhs[4] = mxCreateDoubleMatrix(1, 1, mxREAL);
        tmp = mxGetPr(plhs[4]); *tmp = (double) info[0];
    }
    
    if(nlhs >= 6){
        plhs[5] = mxCreateDoubleMatrix(1, 1, mxREAL);
        tmp = mxGetPr(plhs[5]); *tmp = epsout[0];
    }
    
    if(nlhs >= 7){
        plhs[6] = mxCreateDoubleMatrix(1, 1, mxREAL);
        tmp = mxGetPr(plhs[6]); *tmp = loop[0];
    }
    
    if(spC){
        mxFree(A_indx);
        mxFree(A_pntr);
    }
}
