function [X,ev,Out] = mpmo(Ao,n,k,opts,X)
% Multi-Power Method Outer loop driver for computing
%     k principal eigenpairs of symmetric matrix A
% Usage: [X,ev,Out] = mpmo(Ao,n,k,opts)
%
% Input:
%        Ao -- function handle for symmetric matrix
%         n --  size of Ao
%         k -- number of principal eigenpairs to compute
%         opts -- options (see initialize for details)
%         (Important ones: tol,degs,maxit,nblks,guard)
% Output:
%         X  -- computed k principal eigenvectors
%               (n by k column-orthonormal matrix)
%         ev -- computed k principal eigenvalues
%        Out -- secondary output quantities
%
% Copyright 2013 - 2014 Yin Zhang
% Version 0.8, September 1, 2014
% Version 1,   July 28, 2015: Zaiwen Wen

%% initialization
initialize;  
Aorig = Ao;
kw = k + ceil(guard*k);
S = load('Pcoeffs.mat'); Pc = S.Pc;
maxdeg = min(numel(Pc),opts.maxdeg);

%% continuation parameters %%%%%%%%%%%%%%%%%%%%%%%
% set up tolerance for continuation
tolf = tol; tol0 = opts.tol0; tolInt = opts.tolInt;  
itolf = floor(log10(tolf)); 
if tol > tol0*10^(-tolInt); tolList = tol;
else
    tolList = 10.^(-(abs(log10(tol0)):tolInt:abs(itolf)));
    if tolList(end) > tol; tolList(end) = tol; end
end
tol  = tolList(1); idxtol = 1; ltol = length(tolList);
%tol_deflation = max(tiny,tol^2); %
if tol >= 1e-10; tol_deflation = tiny1;
elseif tol >= 1e-12; tol_deflation = tiny2;
else tol_deflation = max(tiny,tol^2); 
end
tol_deflation_trim = tiny; %max(tiny,tolf^2);

%% initial points
if isempty(X)
    X = randn(n,kw);
else kwmk = kw - size(X,2); %!! better strategy later
    if kwmk > 0; X = [X randn(n,kwmk)]; end
    if kwmk < 0; X = X(:,1:kw); end
end

Qt = []; evt = [];

est_key_eigvals;   % estimate key eigvals           
% possibly trim the other end of spectrum
ell0 = log10((lkw-ln)/(l1-lkw)); ell = ell0;
if istrim, trimming; end
% interval [le,re]
sh = ln; margin = 0; %margin = 0.05;
sm = abs(sh)*margin;
le = ln  - sh - sm;
re = lkw - sh + sm;

%% inner solver parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% set up opts for mpmirr
tol1 = max(eps,min(sqrt(eps),10*tolf^2));
opts1 = struct(...
    'maxit0',maxit0,'maxit1',maxit1,'tol',tol1,'nblks',nblks,...
    'defl',defl,'Qc',[],'evc',[],'Qt',[],'idisp',0,'doARR',doARR);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if idisp
    fprintf('ln: %13.12e, lkw: %13.12e, l1: %13.12e, sh: %13.12e\n', ...
        ln,lkw,l1,sh);
    fprintf('fixdeg: %d, Cheby: %d. l1-ln: %.2e. ell: %.2f',...
        fixdeg, cheby,l1-ln,ell0);
    if ell < ell0, fprintf(' -> %.2f',ell); end
    fprintf('.\n')
end

%% mpm outer loop
maxres = inf; bestres = inf;  ev = []; evb = ev; maxresHist = []; 
ncvf = 0; lk = -inf; cnt = 0;
idxFail = 0;  maxFail = 2;  ires = inf;
tAX = 0; tSU = 0; tRR = 0;
%%%%%%%%%%%%%%%%%%
for it2 = 1:maxit2
    
    %% iterate on mpmirr
    t0 = tic;
    define_Mfun;
    
    if idisp
        fprintf(['\nkw: %3d, nblks: %1d, l1: %2.1e, ln: %2.1e, lk: %2.1e, lkw: %2.1e, ' ...
            'le: %2.1e, re: %2.1e, sh: %2.1e, sm: %2.1e, deg: %2d\n'], ...
            kw, nblks, l1, ln, lk, lkw, le, re, sh, sm, deg);
    end
    
    % call inner solver
    [X,ev,Res,Out1] = feval(opts.innsolver,Aorig,Mfun,X,k,kw,ln,opts1);
    %X = X(:,1:kw);  ev = ev(1:kw);
    tinn = toc(t0); tAX = tAX + Out1.tAX; 
    tSU = tSU + Out1.tSU;  tRR = tRR + Out1.tRR;
    
    %% ----- stop rules and adjust parameters -----
    maxresp = maxres; iresp = ires; ncvfp = ncvf; nxv = length(ev);
    maxres  = max(Res(1:k)); ires = floor(log10(maxres)); 
    maxresHist = [maxresHist, maxres]; %#ok<AGROW>
    
    % how many converged?
    Icv = find(Res(1:k) <= tol_deflation); ncvh = length(Icv);
    
    % fine accuracy convergence
    ncvf = sum(Res(1:k) < tol/10);
    
    Ok = ncvf >= ncvfp || maxres <= maxresp;
    if not(Ok), cnt = cnt + 1; else cnt = 0; end
    rule(1) = maxres < tol*(1 + 9*ncvf/k);
    rule(2) = cnt > 2;
    
    % check for tolf
    %ncvff = sum(Res(1:k) < tolf/10);
    rule(3) = maxres < tolf*(1 + 9*ncvf/k);
    rule(4) = idxFail > maxFail;   
    rule(5) = (iresp == itolf) && (ires == itolf);
    
    % check eigenvalue ratio and increases kw if needed
    rev1 = (ev(min(kw+1, nxv)) - sh) / (ev(k) - sh);
    if rev1 > rgap && maxresp/maxres < 10
        nblks = min(nblks + 1, maxblks);
        opts1.nblks = nblks;
    end
    
    lkp = lk;  lkwp = lkw; l1 = ev(1);  lk  = ev(k); 
    evv = ev(1:kw); lkw = max(min(evv(evv > ln)),ln); %lkw=ev(kw);
    ell0 = log10((lkw-ln)/(l1-lkw)); ell = ell0;
    
    %X = X(:,1:kw);  ev = ev(1:kw);  %Res = Res(1:kw);
    mark = '+';
    if maxres < bestres,
        idxFail = 0;  mark = ' ';
        bestres = maxres; %degb = deg;
        Xb = X; evb = ev; Resb = Res;  
        %l1b = l1; lkb = lk; lkwb = lkw;
    end
    do_disp(mark);

    % termination
    if any(rule(1:2)) || it2 == maxit2 || any(rule(3:5))
        if idxtol == ltol || any(rule(3:5))
            break;
        else
            idxtol = idxtol + 1; tol = tolList(idxtol);
            
            if tol >= 1e-10; tol_deflation = tiny1;
            elseif tol >= 1e-12; tol_deflation = tiny2;
            else tol_deflation = max(tiny,tol^2); 
            end
            
            Icv = find(Res(1:k) <= tol_deflation); ncvh = length(Icv);
            
            cnt = 0;
            %Out1.flag = 1;
            %idxFail = 1;
        end
    end
    
    % reset iteration if res is increased
    if (maxres > bestres && idxFail <= maxFail && it2 > 2)
%         X = Xb; ev = evb; Res = Resb; 
%         maxres = bestres;  
%         l1 = l1b; lk = lkb; lkw = lkwb; 
%         ell0 = log10((lkw-ln)/(l1-lkw)); ell = ell0;
%         deg = deg + 2; 
        idxFail = idxFail + 1;
    end
    
    if defl == 0; Icv = []; end
    
    % get the unconverged index
    iIcv = true(kw,1); iIcv(Icv) = 0; iiIcv = find(iIcv);
    
    %% ----- deflation -----
    if ncvh && defl %deflation
        Qc = X(:,Icv);  
        opts1.Qc = Qc;
        opts1.Qt = Qt;
        opts1.evc = ev(Icv);
    end
    X = X(:,iiIcv);
    
    %% adjust interval and parameters
    if istrim, trimming; end
    %!!! check if trim is executed and kw is part of the trimmed part
    %sh = lkw;
    sh = ln;
    sm = abs(sh)*margin;
    le = ln  - sh - sm;
    re = lkw - sh + sm;
    
    %% adjust polynomial degrees
end % mpm outer loop
%%%%%%%%%%%%%%%%%%%%

%% Output
X = Xb(:,1:k);
ev = evb(1:k);
Out.Res = Resb(1:k);
Out.maxres = maxres;
Out.maxresHist = maxresHist;
Out.rule = rule;
Out.deg = deg;  if fixdeg, Out.deg = maxdeg; end
Out.RR = it2;
Out.tAX = tAX;
Out.tSU = tSU;
Out.tRR = tRR;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%               end of the main code                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% %%%%%%%%%%%%%%%%% %%
%  nested function    %
%%%%%%%%%%%%%%%%%%%%%%%
    function initialize
        %squared = opts.squared;
        maxit0 = opts.maxit(1);
        maxit1 = opts.maxit(2);
        maxit2 = opts.maxit(3);
        cheby  = opts.cheby;
        doARR  = opts.doARR;
        fixdeg = opts.fixdeg;
        nblks  = opts.nblks;
        maxblks = opts.maxblks;
        guard  = opts.guard;
        idisp  = opts.idisp;
        defl   = opts.defl;
        istrim = opts.trim;
        deg    = opts.deg0;
        tol    = opts.tol;
        ln     = opts.ln;
        l1     = opts.l1;
        lkw    = opts.lkw;
        rgap   = opts.rgap;
        tiny   = 1e-14;  tiny1 = 1e-12;  tiny2 = 1e-13;
        tol    = max(tiny,tol); %keep tol >= tiny
    end

%% %%%%%%%%%%%%%%%%% %%
%  nested function    %
%%%%%%%%%%%%%%%%%%%%%%%
    function est_key_eigvals
        %% estimate ln (smallest eigenvalue)
        optse.issym = 1;
        optse.isreal = 1;
        optse.tol = 1e-6;
        if isempty(ln), t = tic;
            ln = eigs(Ao,n,1,'SA',optse);
            tln = toc(t);
        else tln = inf;
        end
        %% estimate lkw and l1 (largest)
        A_ln = @(X)Ao(X)-ln*X;
        Y = X; t = tic;
        for i = 1:deg*3
            Y = A_ln(Y);
            ok = toc(t) > tln/2;
            if ok, break; end
        end
        [Y,~] = qr(Y,0);
        X = A_ln(Y);
        T = X'*Y;
        T = (T + T')/2;
        %%% estimations from T %%%
        if isempty(l1)
            w = eigs(T,2,'BE',optse);
        else
            lb = eigs(T,1,'SA',optse); w = [lb, 0];
        end
        if any(isnan(w)) % not use eigs
            ub = normest(T,1e-3);
            lb = ub-normest(ub*speye(size(T))-T,1e-3);
            w = [lb; ub];
        end
        w = w  + ln; %disp(w)
        
        if isempty(l1)
            lkw = w(1);
            l1  = w(2);
        else
            lkw = max([0.5*lkw,w(1),ln]);
            %lkw = max([w(1),ln]);
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    end

%% %%%%%%%%%%%%%%%%% %%
%  nested function    %
%%%%%%%%%%%%%%%%%%%%%%%
    function trimming
        % precondition an ell-measure
        %if ell0 < 2.5 || l1-ln < 1e+4, return; end
        if ell0 < 2 || l1-ln < 1e+3, return; end
        %if ell0 < 2, return; end
        % trim the other end of spectrum
        t = tic; do_trimming; ell0 = ell;
        ell = log10((lkw-ln)/(l1-lkw));
        if idisp, t = toc(t);
            fprintf('%i trimmed (%.2f sec).\n',ntrim,t);
        end
        % nested function in trimming
        function do_trimming
            %kt = 100;
            kt = max(min(100,round(0.005*n)), 6);
            Xt = randn(n,kt);
            A1 = @(X)l1*X-Ao(X);
            opts1t = struct('maxit0',10,'maxit1',5,...
                'tol',eps,'nblks',4,'idisp',0);
            [Xt,evtt,Rt] = feval(opts.innsolver,A1,A1,Xt,kt,kt,0,opts1t);
            if idisp, fprintf('\n'); fprintf('Rt: %3.2e \n',Rt(1:10)); end
            Icv = Rt <= tol_deflation_trim; ntrim = sum(Icv);
            if ntrim == 0
                opts1t = struct('maxit0',10,'maxit1',5,...
                    'tol',eps,'nblks',2,'idisp',0);
                [Xt,evtt,Rt] = feval(opts.innsolver,A1,A1,Xt,kt,kt,0,opts1t);
                if idisp,fprintf('\n'); fprintf('Rt: %3.2e\n',Rt(1:10));end
                Icv = Rt <= tol_deflation_trim; ntrim = sum(Icv);
            end
            
            if ntrim
                ln = l1 - evtt(min(ntrim+1,length(evtt)));
                Qt = [Qt, Xt(:,Icv)]; evt = [evt, l1 - evtt(Icv)];
                Ao = @(X)Aorig(X)-bsxfun(@times,Qt,evt)*(Qt'*X);
            end
            clear Xt Rt
        end
    end

%% %%%%%%%%%%%%%%%%% %%
%  nested function    %
%%%%%%%%%%%%%%%%%%%%%%%
    function define_Mfun
        Afun = @(X)Ao(X) - sh*X;
        if fixdeg, deg = maxdeg;
        else
            if it2 == 1
                adj = max(-2,min(2,round(ell)));
                deg = deg + adj;
            end
            deg = min(maxdeg,max(2,deg));
            if opts.squared, deg = max(1,round(deg/2)); end

            if ~isempty(evb)
                evcut = opts.evcut; %if maxresp/maxres<10; evcut = 0.8; end
                for degm = 2:maxdeg
                    if cheby
                        emm = MCheby_Value(evb([k,kw])-sh,le,re,degm);
                    else
                        emm =polyvalAX_Value(Pc{degm},evb([k,kw])-sh,le,re);
                    end
                    rrr = emm(2) /emm(1);
                    if rrr < evcut; break; end
                end
                if idisp
                    fprintf('\ndeg: %d, dkw: %3.2e, dk: %3.2e, evcut: %3.2e, ratio: %3.2e\n', ...
                        degm, emm(2), emm(1), evcut, rrr);
                end
                deg = degm;
            end
        end
        
        if cheby || it2 == 1 %use Chebyshev
            Mfun = @(X)MCheby(Afun,X,le,re,deg);
            if ~isempty(ev);
                opts1.evmfun = MCheby_Value(ev(iiIcv)-sh,le,re,deg);
            end
        else
            Mfun = @(X)polyvalAX(Pc{deg},Afun,X,le,re);
            if ~isempty(ev)
                opts1.evmfun = polyvalAX_Value(Pc{deg},ev(iiIcv)-sh,le,re);
            end
        end
    end


%% %%%%%%%%%%%%%%%%% %%
%  nested function    %
%%%%%%%%%%%%%%%%%%%%%%%
    function do_disp(mark)
        if not(idisp), return; end
        dlk  = abs(lk /lkp -1);
        dlkw = abs(lkw/lkwp-1);
        if 0 < 0, dl = dlk; else dl = dlkw; end
        str = 'tol %2.1e it0 %2i(%2i): RRw %4i ncv%4i/%4i ';
        str = [str 'rc %.1e res %.1e %6.1fsec. dl %.1e %s l1-ln: %.2e. ell: %.2f ratio-p: %.2e\n'];
        it0 = Out1.it0; rc = Out1.rc; RRw = Out1.RRw;
        fprintf(str,tol,it0,deg,RRw,ncvh,ncvf,rc,maxres,tinn,dl,mark,l1-ln,ell0, rev1);
        fprintf('inn: %2.1e, tAX: %2.1e, tSU: %2.1e, tRR: %2.1e\n', ...
            tinn, Out1.tAX, Out1.tSU, Out1.tRR);
        
        %semilogy(1:k,Res(1:k)); drawnow, shg; pause(5)
    end
%%%%%%%%%%%
end %mpmo %
%%%%%%%%%%% 

%% %%%%%%%%%%%%%%%%%%%%%%
%  external function    %
%%%%%%%%%%%%%%%%%%%%%%%%%
function X = MCheby(Afun,X,a,b,degree)
% X = MCheby(Afun,X,a,b,degree)
%
if degree == 0; return; end
if degree == 1; X = Afun(X); return; end
%
c0 = (a + b)/(a - b);
c1 = 2 / (b - a);
B = @(X) c0*X + c1*Afun(X);
%
T1 = X; T2 = B(X);
for j = 2:degree
    X = 2*B(T2) - T1;
    if j < degree
        T1 = T2; T2 = X;
    end
end
end


%% %%%%%%%%%%%%%%%%%%%%%%
%  external function    %
%%%%%%%%%%%%%%%%%%%%%%%%%
function Y = polyvalAX(p,A,X,a,b)
% Y = polyvalAX(p,A,X,a,b)

degree = numel(p) - 1;
if degree == 0; Y = p*X; return; end

c0 = (a + b)/(a - b);
c1 = 2 / (b - a);
B = @(X) c0*X + c1*A(X);

Y = p(1)*X; del = eps;
for j = 1:degree
    %Y = B(Y) + p(j+1)*X
    Y = B(Y);
    if abs(p(j+1)) > del
        Y = Y + p(j+1)*X;
    end
end
end

function X = MCheby_Value(X,a,b,degree)
% X = MCheby(Afun,X,a,b,degree)
%
if degree == 0; return; end
if degree == 1; return; end
%
c0 = (a + b)/(a - b);
c1 = 2 / (b - a);
Bx = c0 + c1*X;
%
T1 = 1; T2 = Bx;
for j = 2:degree
    X = 2*Bx.*T2 - T1;
    if j < degree
        T1 = T2; T2 = X;
    end
end
end

function Y = polyvalAX_Value(p,X,a,b)
% Y = polyvalAX(p,A,X,a,b)

degree = numel(p) - 1;
if degree == 0; Y = p*X; return; end

c0 = (a + b)/(a - b);
c1 = 2 / (b - a);
Bx = c0 + c1*X;

Y = p(1); del = eps;
for j = 1:degree
    %Y = B(Y) + p(j+1)*X
    Y = Bx.*Y;
    if abs(p(j+1)) > del
        Y = Y + p(j+1);
    end
end
end
