/* Sparse matrix-dense vector products
 *
 *  X = sfmult(A, B, type)
 *
 *  Inputs:
 *
 *           A - m  x n1 sparse matrix
 *           B - n2 x p  dense  matrix
 *           type - 1 or 2
 *  Outputs:
 *
 *           X = A*B,  if type = 1
 *           X = A'*B, if type = 2
 *
 * Compilation in
 *
 * unix/linux:
 * mex -O -largeArrayDims -output sfmult mkl-sfmult-v1.cpp
 *
 * Author: Zaiwen Wen
 * wenzw@math.pku.edu.cn
 * 2014-5-1
 *
 * reference:
 * https://software.intel.com/en-us/forums/topic/472320
 * https://software.intel.com/en-us/forums/topic/501615
 */


#include "mex.h"
#include "mkl.h"

#if !defined(MX_API_VER) || ( MX_API_VER < 0x07030000 )
typedef int mwIndex;
typedef int mwSize;
#endif

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, 
        const mxArray *prhs[])
{

    size_t m, n, k, type;
    mwIndex *A_Ir, *A_Jc;
    MKL_INT ii, nnz;
    
    // MKL representation
    char NTRANS[] = {'N'}, TTRANS[] = {'T'};
    char A_matdescr[] = {'G', 'L', 'N', 'F'};
    
    MKL_INT A_m[1], B_n[1], A_k[1];
    MKL_INT *A_indx, *A_pntr;
    double  *A, *B, *X, *tt, alpha[1], beta[1];
    MKL_INT ldb[1], ldc[1], i;    
    bool spC;
    
    alpha[0] = 1.0;     beta[0] = 0.0;
    /*default type = 1 */
    type = 1;
    if (nrhs == 3) {
        tt = mxGetPr(prhs[2]); type = (MKL_INT)*tt; }
    if ( (type!= 1) && (type!= 2)) type = 1;
    
    /*
    mkl_set_num_threads(8);
    mkl_domain_set_num_threads(8, MKL_DOMAIN_BLAS);
     */
            
    /*
    int th_id, nthreads;
    nthreads = omp_get_num_threads();
    printf("There are %d threads\n",nthreads);
    */
   
    m = mxGetM(prhs[0]);
    k = mxGetN(prhs[0]);
    n = mxGetN(prhs[1]);
    
      /* Get the dimensions of the matrix input u. */
    spC = mxIsSparse(prhs[0]);
    if(spC) {
        A    = mxGetPr(prhs[0]); 
        A_Ir = mxGetIr(prhs[0]);
        A_Jc = mxGetJc(prhs[0]);
    }
    else    {
        mexErrMsgTxt("Matrix A must be sparse."); return;
    }
    
    B    = mxGetPr(prhs[1]);
    
    A_m[0] = (MKL_INT) m;
    B_n[0] = (MKL_INT) n;
    A_k[0] = (MKL_INT) k;
    
    A_indx = (MKL_INT*) mxMalloc(A_Jc[k]*sizeof(MKL_INT));
    A_pntr = (MKL_INT*) mxMalloc((k)*sizeof(MKL_INT));
    
    // MKL needs 1-based indexing for column-major dense B
    for (i=0; i < (MKL_INT) A_Jc[k]; i++)
        A_indx[i] = (MKL_INT) A_Ir[i] + 1;
    for (i=0; i < A_k[0]+1; i++)
        A_pntr[i] = (MKL_INT) A_Jc[i] + 1;
        
    /*
    nnz = A_Jc[k];
    mexPrintf("\n");
    for(ii = 0; ii < nnz; ii++)
        mexPrintf("row: %d, val: %e, A_indx: %d\n", A_Ir[ii], A[ii], A_indx[ii]);
     mexPrintf("\n");   
    for(ii = 0; ii <= k; ii++)
        mexPrintf("col: %d, Jc: %d, A_pntr: %d\n", ii,  A_Jc[ii], A_pntr[ii]);
    */
    
    
    if(type==1){
        
        if (k != mxGetM(prhs[1]))
            mexErrMsgTxt("Inner matrix dimensions must agree.");

        ldb[0] = A_k[0];
        ldc[0] = A_m[0];
    
        plhs[0] = mxCreateDoubleMatrix(m, n, mxREAL); X = mxGetPr(plhs[0]);
        mkl_dcscmm(NTRANS, A_m, B_n, A_k, alpha, A_matdescr, A,
                A_indx, A_pntr, A_pntr+1, B, ldb, beta, X, ldc);
    
        
    }
    else {
        
        if (m != mxGetM(prhs[1]))
            mexErrMsgTxt("Inner matrix dimensions must agree.");
        
        ldb[0] = A_m[0];
        ldc[0] = A_k[0];
        
        plhs[0] = mxCreateDoubleMatrix(k, n, mxREAL); X = mxGetPr(plhs[0]);
        mkl_dcscmm(TTRANS, A_m, B_n, A_k, alpha, A_matdescr, A,
                A_indx, A_pntr, A_pntr+1, B, ldb, beta, X, ldc);
    }
        
    
    mxFree(A_indx);
    mxFree(A_pntr);
}
